<?php

class BifitKassa
{
	const MODULE_ID = 'bifit.kassa';

	public function __construct()
	{
	}

	public static function __clearString($str)
	{
		$arr = ["\\", "/", "\"", "\n", "\r", "\t", "\x08", "\x0c"];
		$str = str_replace($arr, '', $str);
		$str = preg_replace(
			'/\s\s+/',
			' ',
			preg_replace(
				'/[^ а-яА-ЯёЁ0-9a-zA-Z?@!,.:;\'$№%"*&(){}\[\|\]+-=]/ui',
				'',
				strip_tags(str_replace(" ", " ", $str))
			)
		);

		return trim($str);
	}

	public static function __createGUID($order)
	{
		// Create GUID (Globally Unique Identifier)
		$guid = '';
		$namespace = rand(11111, 99999);
		$uid = uniqid('', true);
		$data = $namespace;
		$data .= $_SERVER['REQUEST_TIME'];
		$data .= $_SERVER['HTTP_USER_AGENT'];
		$data .= $_SERVER['REMOTE_ADDR'];
		$data .= $_SERVER['REMOTE_PORT'];
		$data .= $order['number'];
		$data .= $order['summ'];
		$hash = strtoupper(hash('ripemd128', $uid . $guid . md5($data)));
		$guid =
			substr($hash, 0, 8) .
			'-' .
			substr($hash, 8, 4) .
			'-' .
			substr($hash, 12, 4) .
			'-' .
			substr($hash, 16, 4) .
			'-' .
			substr($hash, 20, 12);

		return strtolower($order['cms'] . '-' . $guid);
	}

	public function __saveLogfile($data, $order_number)
	{
		$file = dirname(__FILE__) . '/logs/log_' . $order_number . '.txt';
		$top =
			"## " .
			date('Y-m-d H:i:s') .
			" | " .
			GetMessage('BIFIT_LOG_BIFIT_KASSA') .
			" | " .
			GetMessage('BIFIT_LOG_ORDER_NUMBER') .
			$order_number;
		$len = mb_strlen($top, "UTF-8");

		for ($i = 0; $i < $len; $i++) {
			$line .= "#";
		}

		$current = file_get_contents($file);
		$content =
			$current . $line . "\n" . $top . "\n" . $line . "\n" . $data . "\n";
		file_put_contents($file, $content);
	}

	protected function __getOrderPayments($orderId, $pmBonusIds = [], $paid = null)
	{
		$payments = [];
		$dbRes = ('\Bitrix\Sale\PaymentCollection')::getList([
			'select' => ['*'],
			'order' => ['ID' => 'ASC'],
			'filter' => ['ORDER_ID' => $orderId]
		]);

		while ($item = $dbRes->fetch()) {
			if (
				is_array($pmBonusIds) &&
				count($pmBonusIds) &&
				in_array($item['PAY_SYSTEM_ID'], $pmBonusIds)
			) {
				continue;
			}
			if ($paid && $item['PAID'] != 'Y') {
				continue;
			}

			$payments[] = $item;
		}

		return $payments;
	}

	function __getPayment($orderId, $pmBonusIds = [])
	{
		$payments = self::__getOrderPayments($orderId);
		$summ = 0;

		foreach ($payments as $item) {
			if (
				is_array($pmBonusIds) &&
				count($pmBonusIds) &&
				!in_array($item['PAY_SYSTEM_ID'], $pmBonusIds)
			) {
				continue;
			}

			$summ += $item['SUM'];
		}

		return $summ;
	}

	function __formatprc($price)
	{
		CModule::IncludeModule("sale");

		$decimals = 2;
		$formatInfo = CCurrencyLang::GetCurrencyFormat(
			COption::GetOptionString("sale", "default_currency", "RUB")
		);

		if (isset($formatInfo['DECIMALS'])) {
			$decimals = intval($formatInfo['DECIMALS']);
		}

		$dec_count = $decimals;

		$price =
			$dec_count < 0
				? round($price, $dec_count)
				: number_format($price, $dec_count, '.', '');

		return $price;
	}

	function __checkInn($inn, $type = 1)
	{
		// check the fullness
		if (isset($inn) && trim($inn) != '') {
			if ($type == 2) {
				// only numbers in the amount of 10 pcs.
				$inn = !preg_match("/^\d{10}$/", trim($inn))
					? GetMessage('BIFIT_WRONG_INN')
					: trim($inn);
			} elseif ($type == 1) {
				// only numbers in the amount of 12 pcs.
				$inn = !preg_match("/^\d{12}$/", trim($inn))
					? GetMessage('BIFIT_WRONG_INN')
					: trim($inn);
			} else {
				// only numbers in the amount of 10 or 12 pieces.
				$inn = !preg_match("/^\d{10}$|^\d{12}$/", trim($inn))
					? GetMessage('BIFIT_WRONG_INN')
					: trim($inn);
			}
		} else {
			// TIN is empty or not found!
			$inn = GetMessage('BIFIT_EMPTY_INN');
		}

		return $inn;
	}

	function __checkPhone($tel)
	{
		// check the fullness
		if (isset($tel) && trim($tel) != '') {
			// remove extra characters
			$phone = str_replace(
				array('+', '(', ')', '-', ' '),
				'',
				trim($tel)
			);

			// get the first digit
			$first = substr($phone, "0", 1);

			// if the number is 10 digits and the first digit is not 7, then add 7
			$phone =
				strlen($phone) == 10 && $first != 7 ? '7' . $phone : $phone;

			// Final inspection
			$phone = !preg_match("/^[0-9]{11}+$/", $phone)
				? str_replace(
					GetMessage('BIFIT_PHONE'),
					GetMessage('BIFIT_PHONE') . ' ' . $phone,
					GetMessage('BIFIT_WRONG_PHONE')
				)
				: $phone;
		} else {
			// The phone is empty or not found!
			$phone = GetMessage('BIFIT_EMPTY_PHONE');
		}

		return $phone;
	}

	function __checkEmail($email)
	{
		// check fullness
		if (isset($email) && trim($email) != '') {
			// record model check
			$email = !preg_match("/^\+\d+$|^\S+@\S+$/", $email)
				? str_replace(
					GetMessage('BIFIT_EMAIL'),
					GetMessage('BIFIT_EMAIL') . ' ' . $email,
					GetMessage('BIFIT_WRONG_EMAIL')
				)
				: $email;
		} else {
			// Email is empty or not found!
			$email = GetMessage('BIFIT_EMPTY_EMAIL');
		}

		return $email;
	}

	function __checkPrice($item_price, $order_subtotal, $order_discount = 0)
	{
		// if the discount is more than zero
		if ($order_discount > 0) {
			// calculate the discount percentage
			$percent = ($order_discount * 100) / $order_subtotal;

			// reduce the price of goods according to the discount
			$item_price = $item_price - $item_price * ($percent / 100);
		}

		return $item_price;
	}

	function __checkResponce($response)
	{
		$mess = '';

		if (
			is_object($response) &&
			($response->type == 'ERROR' || $response->status == 400)
		) {
			$mess = '';
			if ($response->status == 400) {
				$mess .= $response->error . "\n";
				$mess .= $response->message . "\n";
			} else {
				$array = json_decode(json_encode($response), true);
				array_walk_recursive($array, function ($item, $key) use (
					&$mess
				) {
					if ($item != 'ERROR' && trim($item != '')) {
						$mess .= $key . ': ' . $item . "\n";
					}
				});
			}
		} elseif (is_object($response)) {
			$array = json_decode(json_encode($response), true);
			array_walk_recursive($array, function ($item, $key) use (&$mess) {
				if (trim($item != '')) {
					$mess .= $key . ': ' . $item . "\n";
				}
			});
		}

		return $mess;
	}

	function __loadData($url, $headers = false, $data = false, $type = 'POST')
	{
		// initialize the cURL session
		$ch = curl_init($url);

		// assign request type
		if ($type != 'POST') {
			curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $type);
		} else {
			curl_setopt($ch, CURLOPT_POST, true);
		}

		// add data
		if ($data) {
			curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
		}

		// returning the result as a string
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_ENCODING, '');
		curl_setopt($ch, CURLOPT_MAXREDIRS, 10);
		curl_setopt($ch, CURLOPT_TIMEOUT, 0);
		curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
		curl_setopt($ch, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

		// add headers
		if ($headers) {
			curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
		}

		// get the answer
		$output = curl_exec($ch);

		// close the cURL session
		curl_close($ch);

		return $output;
	}

	function __getToken($login, $password)
	{
		$signature = base64_encode(hash('sha256', $password, true));
		$data = 'username=' . $login;
		$data .= '&password=' . $signature;
		$data .= '&client_id=cashdesk-rest-client';
		$data .= '&client_secret=cashdesk-rest-client';
		$data .= '&grant_type=password';
		$url = 'https://kassa.bifit.com/cashdesk-api/v1/oauth/token';
		$headers = array('Content-Type: application/x-www-form-urlencoded');
		$tokenData = json_decode(self::__loadData($url, $headers, $data));

		return $tokenData;
	}

	function __getOrganizations($token)
	{
		$url =
			'https://kassa.bifit.com/cashdesk-api/v1/protected/organizations/list/read';
		$headers = array('Authorization: Bearer ' . $token);
		$orgData = json_decode(self::__loadData($url, $headers));

		return $orgData;
	}

	function __getTradeObjects($org_id, $token)
	{
		$data = 'organization_id=' . $org_id;
		$url =
			'https://kassa.bifit.com/cashdesk-api/v1/protected/trade_objects/list/read';
		$headers = array(
			'Content-Type: application/x-www-form-urlencoded',
			'Authorization: Bearer ' . $token
		);
		$tradeData = json_decode(self::__loadData($url, $headers, $data));

		return $tradeData;
	}

	function __createMainArray($orgData, $token)
	{
		$arr = [];

		foreach ($orgData as $key => $org) {
			$arr[$key] = new stdClass();
			$arr[$key]->id = $org->id;
			$arr[$key]->name = $org->name;
			$arr[$key]->traides = [];

			$tradeData = self::__getTradeObjects($org->id, $token);

			foreach ($tradeData as $k => $trd) {
				$arr[$key]->traides[$k] = new stdClass();
				$arr[$key]->traides[$k]->id = $trd->id;
				$arr[$key]->traides[$k]->name = $trd->name;
			}
		}

		return $arr;
	}

	function __createSelects($array, $name, $org_id = null)
	{
		$html = '<select name="' . $name . '">';

		foreach ($array as $key => $org) {
			if ($name == 'ORGANIZATION') {
				$html .= '<option value="' . $org->id . '">';
				$html .= $org->name;
				$html .= '</option>';
			}

			if ($name == 'TRADE_OBJECT' && $org->id != $org_id) {
				continue;
			}

			if ($name != 'ORGANIZATION') {
				foreach ($org->traides as $k => $trd) {
					$html .= '<option value="' . $trd->id . '">';
					$html .= $trd->name;
					$html .= '</option>';
				}
			}
		}

		$html .= '</select>';

		return $html;
	}

	function __saveReport($order, $receipt_id, $doc, $bifit)
	{
		// display the property of the operation
		$log =
			$order->test != ''
				? '####### ' . GetMessage('BIFIT_TEST_REQUEST') . "\n"
				: '####### ' . GetMessage('BIFIT_BATTLE_REQUEST') . "\n";

		// set the time zone according to the site settings
		$log .= '####### ' . date('Y-m-d H:i:s', time()) . "\n";

		// ANSWER #1

		$log .= '####### ' . GetMessage('BIFIT_ANSWER') . ' #1: ' . "\n";

		// response data # 1
		$token_error = null;

		if (is_object($bifit)) {
			foreach ($bifit as $key => $val) {
				if ($key == 'error') {
					$token_error = 1;
				}
				$log .= $key . ': ' . $val . "\n";
			}
		}

		// RECEIPT

		// generated receipt for sending
		$log .=
			'####### ' .
			GetMessage('BIFIT_CHECK') .
			': ' .
			"\n" .
			$order->receipt .
			"\n";

		if (!$token_error) {
			// IDEMPOTENCY KEY

			$log .=
				'####### ' .
				GetMessage('BIFIT_IDEMPOTENCY_KEY') .
				': ' .
				"\n" .
				$order->idemp_key .
				"\n";

			// ANSWER #2

			$log .= '####### ' . GetMessage('BIFIT_ANSWER') . ' #2: ' . "\n";

			// response data #2
			if (is_object($receipt_id)) {
				$log .= self::__checkResponce($receipt_id);
				$order->receipt_id = GetMessage('BIFIT_NONE_SM');
			} else {
				$order->receipt_id = $receipt_id;
				$log .=
					GetMessage('BIFIT_DOCUMENT_NUMBER') .
					': ' .
					$receipt_id .
					"\n";

				// ANSWER #3

				$log .=
					'####### ' . GetMessage('BIFIT_ANSWER') . ' #3: ' . "\n";

				if (is_object($doc)) {
					$log .= self::__checkResponce($doc);
				}

				// LINK TO THE CHECK

				// generated link to the check
				$log .=
					'####### ' .
					GetMessage('BIFIT_LINK_TO_CHECK') .
					': ' .
					"\n" .
					$order->receipt_link;
			}
		}

		$order->log = $log;

		// SAVE IN DB

		global $DB;

		$connection = ('\Bitrix\Main\Application')::getConnection();
		$sqlHelper = $connection->getSqlHelper();
		$connection = ('\Bitrix\Main\Application')::getConnection();
		$connection->queryExecute(
			"INSERT INTO b_bifit_kassa_reports (
			ORDER_ID, 
			ORDER_NUMBER,
			ORDER_DATE,
			ACTION_DATE,
			RECEIPT_LINK,
			RECEIPT_ID,
			RECEIPT_TYPE,
			LOG
			) VALUES ('" .
				$sqlHelper->forSql($order->id) .
				"', 
			'" .
				$sqlHelper->forSql($order->number) .
				"', 
			'" .
				$sqlHelper->forSql($order->created) .
				"', 
			'" .
				$sqlHelper->forSql($order->changed) .
				"', 
			'" .
				$sqlHelper->forSql($order->receipt_link) .
				"', 
			'" .
				$sqlHelper->forSql($order->receipt_id) .
				"', 
			'" .
				$sqlHelper->forSql($order->receipt_type) .
				"',
			'" .
				$sqlHelper->forSql($order->log) .
				"'
			)"
		);
	}

	function __getOrderTransIds($order_id, $org_id, $token)
	{
		// link
		$url =
			'https://kassa.bifit.com/cashdesk-api/v1/protected/online_orders/details/' .
			$order_id .
			'?organization_id=' .
			$org_id;

		// headers
		$headers = array(
			'Content-Type: application/x-www-form-urlencoded',
			'Authorization: Bearer ' . $token
		);

		// data
		$type = 'GET';
		$data = false;

		// result
		$orderContainer = json_decode(
			self::__loadData($url, $headers, $data, $type)
		);
		$ids = [];

		if (is_object($orderContainer)) {
			$onlineOrderPayments = $orderContainer->onlineOrderPayments;
			if (count($onlineOrderPayments)) {
				foreach ($onlineOrderPayments as $onlinePayment) {
					$ids[] = $onlinePayment->transactionId;
				}
			}
		}

		return $ids;
	}

	function __saveOrder($order)
	{
		Bitrix\Main\Localization\Loc::loadMessages(__FILE__);

		$mid = self::MODULE_ID;

		// log
		$log = '## ' . GetMessage('BIFIT_LOG_GETTING_A_TOKEN') . "\n";

		// Get token
		$data = self::__getToken($order->login, $order->password);
		$token = $data->access_token;

		// log
		if ($token) {
			$log .=
				'## ' .
				GetMessage('BIFIT_LOG_GOT_A_TOKEN') .
				': ' .
				$token .
				"\n";
		} else {
			$log .= '## ' . GetMessage('BIFIT_LOG_TOKEN_ERROR') . "\n";
			$log .=
				'## ' .
				GetMessage('BIFIT_ERROR') .
				': ' .
				$data->error .
				"\n" .
				'## ' .
				GetMessage('BIFIT_ERROR_DESC') .
				': ' .
				$data->error_description;

			self::__saveLogfile($log, $order->number);

			return false;
		}

		// log
		$log .=
			'## ' . GetMessage('BIFIT_LOG_CHECK_AVAILABILITY_PRODUCT') . "\n";

		// parameters for ordering
		$onlineOrderItems = '';

		// form order data
		foreach ($order->products as $key_id => $product) {
			// address for POST request
			$url =
				'https://kassa.bifit.com/cashdesk-api/v1/protected/nomenclatures/external_id/' .
				$product->id .
				'?organization_id=' .
				$order->org_id;

			// link data
			$data = false;

			// headers
			$headers = array(
				'Content-Type: application/x-www-form-urlencoded',
				'Authorization: Bearer ' . $token
			);

			$prod = json_decode(self::__loadData($url, $headers, $data, 'GET'));

			if ($prod == null) {
				// CREATE A NEW PRODUCT

				// log
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_PRODUCT_VS_ID') .
					$product->id .
					' ' .
					GetMessage('BIFIT_LOG_NOT_FOUND_IN_BIFIT_DATABASE') .
					"\n";
				$log .= '## ' . GetMessage('BIFIT_LOG_ADD_NEW_PRODUCT') . "\n";

				// address for POST request
				$url =
					'https://kassa.bifit.com/cashdesk-api/v1/protected/nomenclatures';

				// link data
				$data = '{';
				$data .= '"item":{';
				$data .=
					'"vatValue":' .
					($order->vat_mode == 1
						? $order->vat_value
						: $product->tax_value) .
					',';
				$data .= '"vendorCode":"' . $product->code . '",';
				$data .= '"name":"' . $product->name . '",';
				$data .= '"externalId":"' . $product->id . '",';
				$data .= '"sellingPrice":' . $product->price;
				$data .= '},';
				$data .= '"organizationId":"' . $order->org_id . '"';
				$data .= '}';

				// headers
				$headers = array(
					'Content-Type: application/json',
					'Authorization: Bearer ' . $token
				);

				// log
				$log .= GetMessage('BIFIT_LOG_DATA_FOR_REQUEST') . ':' . "\n";
				$log .= $data . "\n";

				// get data from request
				$tov = json_decode(self::__loadData($url, $headers, $data));

				// log
				if (!is_object($tov) && !is_array($tov)) {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_SUCC_ADD_NEW_PRODUCT') .
						$tov .
						"\n";
				} else {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_ERROR_WHEN_ADDED_NEW_PRODUCT') .
						':';
					$log .= self::__checkResponce($tov) . "\n";
				}
			} else {
				$tov = $prod->id;

				// log
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_PRODUCT_VS_ID') .
					$product->id .
					' ' .
					GetMessage('BIFIT_LOG_FOUND_IN_BIFIT_DATABASE_VS_ID') .
					$tov .
					"\n";
			}

			// display product data in the order
			$onlineOrderItems .= $onlineOrderItems != '' ? ', ' : '';
			$onlineOrderItems .= '{';
			$onlineOrderItems .= '"nomenclatureId":"' . $tov . '",';
			$onlineOrderItems .= '"paymentSubject":"PRODUCT",';
			$onlineOrderItems .= '"name":"' . $product->name . '",';
			$onlineOrderItems .= '"price":' . $product->price . ',';
			$onlineOrderItems .=
				'"vatValue":' .
				($order->vat_mode == 1
					? $order->vat_value
					: $product->tax_value) .
				',';
			$onlineOrderItems .= '"quantity":' . $product->count . ',';
			$onlineOrderItems .= '"vendorCode":"' . $product->code . '"';
			$onlineOrderItems .= '}';
		}

		// delivery details
		if ($order->delivery_cost > 0) {
			// address for POST request
			$url =
				'https://kassa.bifit.com/cashdesk-api/v1/protected/nomenclatures/external_id/' .
				'delivery_' .
				$order->delivery_id .
				'?organization_id=' .
				$order->org_id;

			// link data
			$data = false;

			// headers
			$headers = array(
				'Content-Type: application/x-www-form-urlencoded',
				'Authorization: Bearer ' . $token
			);

			$delivery = json_decode(
				self::__loadData($url, $headers, $data, 'GET')
			);

			// log
			$log .=
				'## ' . GetMessage('BIFIT_LOG_CHECK_DELIVERY_SERVICE') . "\n";

			if ($delivery == null) {
				// CREATE A NEW SERVICE

				// log
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_DELIVERY_SERVICE_NOT_FOUND') .
					"\n";
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_ADD_NEW_DELIVERY_SERVICE') .
					"\n";

				// address for POST request
				$url =
					'https://kassa.bifit.com/cashdesk-api/v1/protected/nomenclatures';

				// link data
				$data = '{';
				$data .= '"item":{';
				$data .=
					'"vatValue":' .
					($order->vat_mode == 1
						? $order->vat_value
						: $order->delivery_tax_value) .
					',';
				$data .= '"name":"' . GetMessage('BIFIT_DELIVERY') . '",';
				$data .= '"externalId":"delivery_' . $order->delivery_id . '",';
				$data .= '"sellingPrice":' . $order->delivery_cost;
				$data .= '},';
				$data .= '"organizationId":"' . $order->org_id . '"';
				$data .= '}';

				// headers
				$headers = array(
					'Content-Type: application/json',
					'Authorization: Bearer ' . $token
				);

				// log
				$log .= GetMessage('BIFIT_LOG_DATA_FOR_REQUEST') . ':' . "\n";
				$log .= $data . "\n";

				// get data from request
				$tov = json_decode(self::__loadData($url, $headers, $data));

				// log
				if (!is_object($tov) && !is_array($tov)) {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_SUCC_ADD_NEW_DELIVERY_SERVICE') .
						$tov .
						"\n";
				} else {
					$log .=
						'## ' .
						GetMessage(
							'BIFIT_LOG_ERROR_WHEN_ADDED_NEW_DELIVERY_SERVICE'
						) .
						"\n";
					$log .= self::__checkResponce($tov) . "\n";
				}
			} else {
				$tov = $delivery->id;

				// log
				$log .=
					'## ' .
					GetMessage(
						'BIFIT_LOG_DELIVERY_SERVICE_ALREADY_EXIST_VS_ID'
					) .
					$tov .
					"\n";
			}

			// display data of delivery method in order
			$onlineOrderItems .= $onlineOrderItems != '' ? ',' : '';
			$onlineOrderItems .= '{';
			$onlineOrderItems .= '"nomenclatureId":"' . $tov . '",';
			$onlineOrderItems .= '"paymentSubject":"SERVICE",';
			$onlineOrderItems .=
				'"name":"' . GetMessage('BIFIT_DELIVERY') . '",';
			$onlineOrderItems .= '"price":' . $order->delivery_cost . ',';
			$onlineOrderItems .=
				'"vatValue":' .
				($order->vat_mode == 1
					? $order->vat_value
					: $order->delivery_tax_value) .
				',';
			$onlineOrderItems .= '"quantity":1';
			$onlineOrderItems .= '}';
		}

		// CREATE A USER IN THE CASH

		// address for POST request
		$url = 'https://kassa.bifit.com/cashdesk-api/v1/protected/clients';

		// headers
		$headers = array(
			'Content-Type: application/json',
			'Authorization: Bearer ' . $token
		);

		// link data
		$data = '{';
		$data .= '"organizationId":"' . $order->org_id . '",';
		$data .= '"address":"' . $order->user_address . '",';
		$data .= '"phone":"' . $order->user_phone . '",';
		$data .= '"lastName":"' . $order->user_last_name . '",';
		$data .= '"firstName":"' . $order->user_name . '",';
		$data .= '"email":"' . $order->user_email . '"';
		$data .= '}';

		// log
		$log .= '## ' . GetMessage('BIFIT_LOG_CHECK_CLIENT') . "\n";
		$log .= '## ' . GetMessage('BIFIT_LOG_DATA_FOR_REQUEST') . ':' . "\n";
		$log .= $data . "\n";

		// trying to add a new client
		$client = json_decode(self::__loadData($url, $headers, $data));

		// such a client is already in the database
		if (is_object($client) && $client->type == "ERROR") {
			// address for POST request
			$url =
				'https://kassa.bifit.com/cashdesk-api/v1/protected/clients/list/read?organization_id=' .
				$order->org_id;

			// link data
			$data = false;

			// headers
			$headers = array(
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token
			);

			// get all clients
			$clients = json_decode(self::__loadData($url, $headers, $data));

			foreach ($clients as $item) {
				// if some user data matches
				if ($item->phone == $order->user_phone) {
					$client = $item->id;
					$clientAddress = $item->address;
					$clientLastName = $item->lastName;
					$clientFirstName = $item->firstName;
					$clientPatronymic = $item->patronymic;
					$clientOrganizationInn = $item->organizationInn;
					$clientOrganizationName = $item->organizationName;
					$clientEmail = $item->email;
					$clientComment = $item->comment;
					$clientStatus = $item->status;
					$clientLoyalty = $item->loyalty;
					break;
				}
			}

			$cur =
				'{"address":"' .
				$clientAddress .
				'","lastName":"' .
				$clientLastName .
				'","firstName":"' .
				$clientFirstName .
				'","email":"' .
				$clientEmail .
				'"}';

			// log
			$log .=
				'## ' .
				GetMessage('BIFIT_LOG_CLIENT_ALREADY_EXIST') .
				$client .
				"\n";
			$log .=
				'## ' .
				GetMessage('BIFIT_LOG_CLIENT_DATA_CHECK') .
				"\n" .
				$cur .
				"\n";

			if (
				($order->user_email != '' &&
					$clientEmail != $order->user_email) ||
				($order->user_name != '' &&
					$clientFirstName != $order->user_name) ||
				($order->user_last_name != '' &&
					$clientLastName != $order->user_last_name)
			) {
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_CLIENT_DATA_BAD_START_UPDATE') .
					"\n";

				// UPDATE CLIENT DATA

				// address for PUT request
				$url =
					'https://kassa.bifit.com/cashdesk-api/v1/protected/clients/' .
					$client;

				// link data
				$data = '{';
				$data .= '"id":"' . $client . '",';
				$data .= '"organizationId":"' . $order->org_id . '",';
				$data .=
					'"address":"' .
					($order->user_address != '' &&
					$order->user_address != $clientAddress
						? $order->user_address
						: $clientAddress) .
					'",';
				$data .= '"phone":"' . $order->user_phone . '",';
				$data .=
					'"lastName":"' .
					($order->user_last_name != '' &&
					$order->user_last_name != $clientLastName
						? $order->user_last_name
						: $clientLastName) .
					'",';
				$data .=
					'"firstName":"' .
					($order->user_name != '' &&
					$order->user_name != $clientFirstName
						? $order->user_name
						: $clientFirstName) .
					'",';
				$data .= '"patronymic":"' . $clientPatronymic . '",';
				$data .= '"organizationInn":"' . $clientOrganizationInn . '",';
				$data .=
					'"organizationName":"' . $clientOrganizationName . '",';
				$data .=
					'"email":"' .
					($order->user_email != '' &&
					$order->user_email != $clientEmail
						? $order->user_email
						: $clientEmail) .
					'",';
				$data .=
					'"comment":"' .
					($order->user_comment != '' &&
					$order->user_comment != $clientComment
						? $order->user_comment
						: $clientComment) .
					'",';
				$data .= '"status":"' . $clientStatus . '",';
				$data .= '"loyalty":"' . $clientLoyalty . '"';
				$data .= '}';

				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_CLIENT_DATA_UPDATE_DATA') .
					"\n" .
					$data .
					"\n";

				// headers
				$headers = array(
					'Content-Type: application/json',
					'Authorization: Bearer ' . $token
				);

				// put data
				$result = json_decode(
					self::__loadData($url, $headers, $data, 'PUT')
				);

				if (!is_object($result) && !is_array($result) && $result) {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_CLIENT_DATA_UPDATE_GOOD') .
						"\n";
				} else {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_CLIENT_DATA_UPDATE_ERROR') .
						"\n" .
						self::__checkResponce($result);
				}
			} else {
				$log .= '## ' . GetMessage('BIFIT_LOG_CLIENT_DATA_OK') . "\n";
			}
		} else {
			// log
			$log .=
				'## ' .
				GetMessage('BIFIT_LOG_ADDED_NEW_CLIENT_VS_ID') .
				$client .
				"\n";
		}

		// SENDING AN ORDER TO THE CASHIER

		// checking order availability
		// address for GET request
		$url =
			'https://kassa.bifit.com/cashdesk-api/v1/protected/online_orders/list/read/' .
			$client .
			'?organization_id=' .
			$order->org_id;

		// link data
		$data = false;

		// headers
		$headers = array(
			'Content-Type: application/x-www-form-urlencoded',
			'Authorization: Bearer ' . $token
		);

		// get orders
		$orders = json_decode(self::__loadData($url, $headers, $data, 'GET'));

		// params
		$order_id = null;
		$delivery_type = '';
		$order_deleted = null;

		// log
		$log .= '## ' . GetMessage('BIFIT_LOG_CHECK_CURR_ORDER') . "\n";

		// trans
		$l = [
			'COURIER' => GetMessage('BIFIT_LOG_TYPE_COURIER'),
			'SELF' => GetMessage('BIFIT_LOG_TYPE_SELF')
		];

		// search order by id
		foreach ($orders as $key => $item) {
			if ($item->externalId == $order->number) {
				$order_id = $item->id;
				$delivery_type = strtoupper($item->deliveryType);

				// log
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_THIS_ORDER_ALREADY_EXIST') .
					$order_id .
					'. ' .
					GetMessage('BIFIT_LOG_DELIVERY_TYPE') .
					$l[$delivery_type] .
					'.' .
					"\n";

				break;
			}
		}

		// log
		if (!$order_id) {
			$log .= '## ' . GetMessage('BIFIT_LOG_CURR_ORDER_NOT_FOUND') . "\n";
		}

		// type of delivery
		$deliveryType = 'SERVICE';
		if (
			isset($order->courier) &&
			in_array($order->delivery_id, $order->courier)
		) {
			$deliveryType = 'COURIER';
		} elseif (
			isset($order->self) &&
			in_array($order->delivery_id, $order->self)
		) {
			$deliveryType = 'SELF';
		}

		// assign "is_update"
		$is_update =
			isset($order_id) &&
			isset($order->refused_statuses) &&
			in_array($order->status_id, $order->refused_statuses) &&
			$order->use_refused
				? 1
				: null;

		if ($deliveryType == 'SERVICE') {
			// log
			$txt = !$order_id
				? GetMessage('BIFIT_LOG_FAILED_TO_SAVE_CURR_ORDER')
				: GetMessage('BIFIT_LOG_FAILED_TO_UPDATE_CURR_ORDER');
			$txt .= ' ' . GetMessage('BIFIT_LOG_FAILED_TO_SERVICE_TEXT');
			$log .= $txt != '' ? '## ' . $txt . "\n" : '';
		} elseif (!$order_id || $is_update) {
			// SAVE || UPDATE order

			// address for request
			$url =
				'https://kassa.bifit.com/cashdesk-api/v1/protected/online_orders';

			// headers
			$headers = array(
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token
			);

			// data
			if ($is_update) {
				// log
				$log .=
					'## ' . GetMessage('BIFIT_LOG_START_ORDER_UPDATE') . "\n";

				$url .= '/' . $order_id . '/status';
				$type = 'PUT';
				$statusType = $delivery_type == 'SELF' ? 'REFUSED' : 'DELETED';

				// Basic data
				$data = '{';
				$data .= '"onlineOrderId":"' . $order_id . '",';
				$data .= '"statusType":"' . $statusType . '",';
				$data .= '"statusDate":0,';
				$data .= '"statusInfo":""';
				$data .= '}';
			} elseif (!$order_id) {
				// log
				$log .= '## ' . GetMessage('BIFIT_LOG_START_NEW_ORDER') . "\n";

				// data
				$type = 'POST';
				$statusType =
					$deliveryType == 'SELF'
						? 'ACCEPTED'
						: strtoupper($order->new_status);

				// timestamp
				$timestamp = strtotime($order->created);
				$orderTime =
					strlen($timestamp) == 10 ? $timestamp * 1000 : $timestamp;

				// Basic data
				$onlineOrder = '{';
				$onlineOrder .= '"organizationId":"' . $order->org_id . '",';
				$onlineOrder .= '"tradeObjectId":"' . $order->trade_id . '",';
				$onlineOrder .=
					'"taxSystem":"' . strtoupper($order->tax_system) . '",';
				$onlineOrder .= '"externalId":"' . $order->number . '",';
				$onlineOrder .= '"deliveryType":"' . $deliveryType . '",';
				$onlineOrder .=
					'"deliveryAmount":"' . $order->delivery_cost . '",';
				$onlineOrder .= '"paid":false,';
				$onlineOrder .= '"orderTime":' . $orderTime . ',';
				$onlineOrder .= '"clientId":' . $client . ',';
				$onlineOrder .=
					'"address":"' . addslashes($order->user_address) . '",';
				$onlineOrder .=
					'"comment":"' .
					htmlspecialchars(
						trim($order->user_comment . ' ' . $order->comment)
					) .
					'",';
				$onlineOrder .=
					'"responsiblePersonLogin":"' . $order->login . '",';
				$onlineOrder .= '"totalAmount":' . $order->total . ',';
				$onlineOrder .= '"currentStatusType":"' . $statusType . '",';
				$onlineOrder .= '"currentStatusTime":0';
				$onlineOrder .= '}';

				$data = '{';
				$data .= '"onlineOrder":' . $onlineOrder . ',';
				$data .= '"onlineOrderItems":[' . $onlineOrderItems . ']';
				//$data .= ',"onlineOrderPayments":[]';
				$data .= '}';
			}

			// log
			$log .=
				'## ' .
				GetMessage('BIFIT_LOG_ORDER_DATA_FOR_REQUEST') .
				':' .
				"\n";
			$log .= $data . "\n";

			// get the ID of the document being processed (id)
			$order_id = json_decode(
				self::__loadData($url, $headers, $data, $type)
			);

			// log
			if (!is_object($order_id) && !is_array($order_id)) {
				if ($is_update && $statusType == 'DELETED') {
					$order_deleted = true;
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_SUCC_ORDER_DEL') .
						$order_id .
						"\n";
				} elseif ($is_update && $statusType == 'REFUSED') {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_SUCC_ORDER_REFUSED') .
						$order_id .
						"\n";
				} else {
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_SUCC_ORDER_SAVE') .
						$order_id .
						"\n";
				}
			} else {
				$log .=
					'## ' .
					GetMessage('BIFIT_LOG_ERR_IN_SAVE_ORDER_PROC') .
					':' .
					"\n";
				$log .= self::__checkResponce($order_id) . "\n";
			}
		}

		// SEND PAYMENT INFORMATION

		if (
			is_numeric($order_id) &&
			!$order_deleted &&
			in_array(strtoupper($order->calc_method), [
				'PREPAY_FULL',
				'PREPAY_PARTIAL',
				'FULL_PAY'
			]) &&
			($order->paid == 'Y' || $order->sum_paid > 0)
		) {
			// get order payments
			$orderTransIds = self::__getOrderTransIds(
				$order_id,
				$order->org_id,
				$token
			);

			// params
			$calc_method = strtoupper($order->calc_method);
			$total =
				$order->sum_paid && $order->sum_paid > 0 && $order->total > 0
					? $order->sum_paid
					: $order->total;
			$ok_payments = $order->ok_payments;
			$k = (int)(count($ok_payments) - 1);
			$last_ok_pm = $ok_payments[$k];

			if (count($order->payments)) {
				if (count($order->payments) > 1) {
					$calc_method = 'PREPAY_PARTIAL';
				}

				foreach ($order->payments as $item) {
					if ($item['PAID'] != 'Y') {
						continue;
					}
					if (
						$order->bonus &&
						in_array(
							$item['PAY_SYSTEM_ID'],
							$order->pm_bonus_ids
						) &&
						$order->total > 0
					) {
						continue;
					}

					// check item price
					$item_total =
						$order->total == 0 ? $order->total : $item['SUM'];

					// correct item summ
					if (
						$order->paid == 'Y' && 
						$order->total > 0 && 
						$order->total != $order->sum_paid && 
						$last_ok_pm['XML_ID'] == $item['XML_ID']
					) {
						$item_total = $item['SUM'] + (float)($order->total - $order->sum_paid);
					}

					// log
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_START_PAY_DATA_PROC') .
						"\n";

					// address for request
					$url =
						'https://kassa.bifit.com/cashdesk-api/v1/protected/online_orders/payments';

					// headers
					$headers = array(
						'Content-Type: application/json',
						'Authorization: Bearer ' . $token
					);

					// data
					$type = 'POST';
					$data = '{';
					$data .=
						'"onlineOrderExternalId":"' . $order->number . '",';

					if (
						!is_object($order_id) &&
						!is_array($order_id) &&
						is_numeric($order_id)
					) {
						$data .= '"onlineOrderId":' . (int) $order_id . ',';
					}

					$data .= '"transactionId":"' . $item['XML_ID'] . '",';
					$data .= '"organizationId":"' . $order->org_id . '",';
					$data .= '"calculationMethod":"' . $calc_method . '",';
					$data .= '"total":' . $item_total;
					$data .= '}';

					// log
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_PAY_DATA_FOR_REQUEST') .
						':' .
						"\n";
					$log .= $data . "\n";
					$log .=
						'## ' . GetMessage('BIFIT_LOG_PAY_DATA_CHECK') . "\n";

					// check exist
					if (
						count($orderTransIds) &&
						in_array($item['XML_ID'], $orderTransIds)
					) {
						// log
						$log .=
							'## ' .
							GetMessage('BIFIT_LOG_PAY_DATA_CHECK_EXIST') .
							"\n";
						continue;
					}

					// log
					$log .=
						'## ' .
						GetMessage('BIFIT_LOG_PAY_DATA_CHECK_NO_EXIST') .
						"\n";

					// get the ID of the document being processed (id)
					$response = json_decode(
						self::__loadData($url, $headers, $data, $type)
					);

					// log
					if (!is_object($response) && !is_array($response)) {
						$log .=
							'## ' .
							GetMessage('BIFIT_LOG_SUCC_SEND_PAY_DATA') .
							$response .
							"\n";
					} else {
						$log .=
							'## ' .
							GetMessage('BIFIT_LOG_ERR_WHEN_SEND_PAY_DATA') .
							':' .
							"\n";
						$log .= self::__checkResponce($response) . "\n";
					}

					if ($order->total == 0) {
						break;
					}
				}
			}
		}

		// save log
		self::__saveLogfile($log, $order->number);
	}

	function __doFiscalize($order)
	{
		Bitrix\Main\Localization\Loc::loadMessages(__FILE__);

		$mid = self::MODULE_ID;

		// 1. AUTHORIZATION
		//    process necessary to connect a client to fiscal processing

		// address for POST request
		$url =
			'https://fp' .
			$order->test .
			'.bifit.com/processing-api/oauth/token';

		// link data
		$data = 'token=' . $order->token;
		$data .= '&client_id=processing-connector-token';
		$data .= '&client_secret=processing-connector-token';
		$data .= '&grant_type=token';

		// headers
		$headers = array('Content-Type: application/x-www-form-urlencoded');

		// decode the received json response
		$bifit = json_decode(self::__loadData($url, $headers, $data));

		// 2. FORMATION OF THE CONTENT OF A CHECK
		//    formation of information for further sending for fiscalization

		// params
		$calc_method = strtoupper($order->calc_method);
		$payment_type = strtoupper($order->payment_type);
		$payments = '"' . $payment_type . '":' . $order->total;

		// avans
		if (
			$order->avans &&
			count($order->payments) > 1 &&
			$order->total > 0 &&
			$order->total <= $order->sum_paid
		) {
			$i = 0;
			$payments = '';
			foreach ($order->payments as $item) {
				if ($item['PAID'] != 'Y') {
					continue;
				}
				if (
					$order->bonus &&
					in_array($item['PAY_SYSTEM_ID'], $order->pm_bonus_ids) &&
					$order->total > $order->bonus_summ
				) {
					continue;
				}

				$payment_type =
					$i == 0 ? 'PREPAY' : strtoupper($order->payment_type);
				$payments .= $payments != '' ? ',' : '';
				$payments .= '"' . $payment_type . '":' . $item['SUM'];
				$i++;
			}
		}

		// check parameters
		$items = '';

		// generate check data
		foreach ($order->products as $key_id => $product) {
			// display the product data in the check
			$items .= $items != '' ? ',' : '';
			$items .= '{';
			$items .= '"calculationMethod":"' . strtoupper($calc_method) . '",';
			$items .= '"paymentSubject":"PRODUCT",';
			$items .= '"name":"' . $product->name . '",';
			$items .= '"price":' . $product->price . ',';
			$items .= '"quantity":' . $product->count . ',';
			$items .=
				$order->vat_mode == 1
					? '"vat":"' . strtoupper($order->vat) . '",'
					: '"vat":"' . strtoupper($product->tax) . '",';
			$items .= '"total":' . (float) ($product->price * $product->count);
			$items .= '}';
		}

		// delivery
		if ($order->delivery_cost > 0) {
			// display the data of the delivery method in the check
			$items .= $items != '' ? ',' : '';
			$items .= '{';
			$items .= '"calculationMethod":"' . strtoupper($calc_method) . '",';
			$items .= '"paymentSubject":"SERVICE",';
			$items .= '"name":"' . GetMessage('BIFIT_DELIVERY') . '",';
			$items .= '"price":' . $order->delivery_cost . ',';
			$items .= '"quantity":1,';
			$items .=
				$order->vat_mode == 1
					? '"vat":"' . strtoupper($order->vat) . '",'
					: '"vat":"' . strtoupper($order->delivery_tax) . '",';
			$items .= '"total":' . $order->delivery_cost;
			$items .= '}';
		}

		// form a check
		$receipt = '{';
		$receipt .= '"type":"' . strtoupper($order->receipt_type) . '",';
		$receipt .= '"taxSystem":"' . strtoupper($order->tax_system) . '",';
		$receipt .= '"cashier":{';
		$receipt .= '"name":"' . $order->cashier_name . '",';
		$receipt .= '"inn":"' . self::__checkInn($order->cashier_inn, 1) . '"';
		$receipt .= '},';
		$receipt .= '"client":{';
		$receipt .= $order->fio
			? '"name":"' .
				trim($order->user_name . ' ' . $order->user_last_name) .
				'",'
			: '';
		$receipt .= '"address":"' . $order->address . '"';
		$receipt .= '},';
		$receipt .= '"paymentAddress":"' . $order->payment_address . '",';
		$receipt .= '"items":[' . $items . '],';
		$receipt .= '"total":' . $order->total . ',';
		$receipt .= '"payments":{' . $payments . '}';
		$receipt .= '}';

		// 3. SENDING A CHECK TO FISCAL PROCESSING
		//    transfer of the content of the check to fiscal processing for further fiscalization

		// address for POST request
		$url =
			'https://fp' .
			$order->test .
			'.bifit.com/processing-api/protected/documents/registration/receipts';

		// check data
		$data = $receipt;

		// Idempotency-Key
		$o = [];
		$o['number'] = $order->number;
		$o['summ'] = $order->total;
		$o['cms'] = 'Bitrix';
		$idemp_key = self::__createGUID($o);

		// headers
		$headers = array(
			'Content-Type: application/json',
			'Authorization: Bearer ' . $bifit->access_token,
			'Idempotency-Key: ' . $idemp_key
		);

		// get the ID of the document being processed (id)
		$receipt_id = json_decode(self::__loadData($url, $headers, $data));

		// 4. RECEIVING DOCUMENT BY ID
		//    obtaining a document with fiscal characteristics by its number

		// options
		$doc = '';

		if (!is_object($receipt_id)) {
			// link for GET request
			$url =
				'https://fp' .
				$order->test .
				'.bifit.com/processing-api/protected/documents/' .
				$receipt_id;

			// headers
			$headers = array('Authorization: Bearer ' . $bifit->access_token);

			// we get the document
			$doc = json_decode(self::__loadData($url, $headers, false, 'GET'));
		}

		// LINK TO THE CHECK

		// we form a link to the check
		$receipt_link = !is_object($receipt_id)
			? 'https://fp' .
				$order->test .
				'.bifit.com/processing-api/receipts/' .
				$idemp_key
			: 'XXX';

		// OPERATION HISTORY

		// update data
		$order->idemp_key = $idemp_key;
		$order->receipt = $receipt;
		$order->receipt_link = $receipt_link;

		// save the operation to history
		self::__saveReport($order, $receipt_id, $doc, $bifit);
	}

	public static function __getOrderData($order)
	{
		// params
		$mid = self::MODULE_ID;
		$user_phone = null;
		$codes = array('LOCATION', 'ZIP', 'CITY', 'ADDRESS');
		$uaa = [];
		$op = CSaleOrderPropsValue::GetList(
			array("SORT" => "ASC"),
			array("ORDER_ID" => $order['ID'])
		);
		while ($aop = $op->Fetch()) {
			if ($aop['CODE'] == "EMAIL" && $aop['VALUE'] != '') {
				$user_email = $arOrderProps['VALUE'];
			}
			if ($aop['CODE'] == "PHONE" && $aop['VALUE'] != '') {
				$user_phone = $aop['VALUE'];
			}
			if (in_array($aop['CODE'], $codes) && $aop['VALUE'] != '') {
				if ($aop['CODE'] == 'LOCATION') {
					$res = ('\Bitrix\Sale\Location\LocationTable')::getList(
						array(
							'filter' => array(
								'=CITY_ID' => $aop['VALUE'],
								'=PARENTS.NAME.LANGUAGE_ID' => LANGUAGE_ID,
								'=PARENTS.TYPE.NAME.LANGUAGE_ID' => LANGUAGE_ID
							),
							'select' => array(
								'I_ID' => 'PARENTS.ID',
								'I_NAME_RU' => 'PARENTS.NAME.NAME',
								'I_TYPE_CODE' => 'PARENTS.TYPE.CODE',
								'I_TYPE_NAME_RU' => 'PARENTS.TYPE.NAME.NAME'
							),
							'order' => array(
								'PARENTS.DEPTH_LEVEL' => 'asc'
							)
						)
					);
					while ($item = $res->fetch()) {
						if (!in_array($item['I_NAME_RU'], $uaa)) {
							$uaa[] = $item['I_NAME_RU'];
						}
					}
				} else {
					$uaa[] = $aop['VALUE'];
				}
			}
		}
		$user_address = count($uaa) ? implode(', ', $uaa) . '.' : null;
		$address_type = COption::GetOptionString($mid, 'address_type', 'email');

		// check bonus discounts
		$bonus_summ = 0;
		if (COption::GetOptionString($mid, 'bonus')) {
			$pm_bonus_ids = json_decode(
				COption::GetOptionString($mid, 'pm_bonus_ids')
			);
			$bonus_summ = self::__getPayment($order['ID'], $pm_bonus_ids);
		}

		// discounts summ
		$order_discount = $bonus_summ + abs($order['DISCOUNT_VALUE']);

		// products
		$products = [];
		$cart = CSaleBasket::GetList(
			array("ID" => "ASC"),
			array("ORDER_ID" => $order['ID']),
			false,
			false
		);
		$cart_new_summ = 0;
		$k = 0;
		while ($product = $cart->GetNext()) {
			$tax_value =
				$product['VAT_INCLUDED'] == 'Y' ? 0 : $order['TAX_VALUE'];
			$price = self::__formatprc(
				self::__checkPrice(
					$product['PRICE'],
					$order['PRICE'] + $tax_value,
					$order_discount
				)
			);
			$name = preg_replace(
				'/ +/',
				' ',
				trim(
					preg_replace(
						'/[^\p{L}0-9 ]/iu',
						' ',
						html_entity_decode($product['NAME'])
					)
				)
			);
			$ar_res = CCatalogProduct::GetByIDEx($product['PRODUCT_ID']);
			$articul = $ar_res['PROPERTIES']['ARTNUMBER']['VALUE'];

			$products[$k] = new stdClass();
			$products[$k]->id = $product['PRODUCT_ID'];
			$products[$k]->name = $name;
			$products[$k]->price = $price;
			$products[$k]->count = $product['QUANTITY'];
			$products[$k]->tax =
				(float) $product['VAT_RATE'] > 0
					? 'VAT_' . (int) ($product['VAT_RATE'] * 100)
					: 'WITHOUT_VAT';
			$products[$k]->tax_value =
				(float) $product['VAT_RATE'] > 0
					? (int) ($product['VAT_RATE'] * 100)
					: 0;
			$products[$k]->code = $articul;

			$cart_new_summ += $price * $product['QUANTITY'];
			$k++;
		}

		// correct delivery price
		$order_delivery_price = 0;
		if ($order['PRICE_DELIVERY'] > 0) {
			$order_delivery_price = self::__formatprc(
				self::__checkPrice(
					$order['PRICE_DELIVERY'],
					$order['PRICE'],
					$order_discount
				)
			);
		}

		// correct total summ
		$correction = null;
		$new_order_price = $cart_new_summ + $order_delivery_price;
		$order_price = $order['PRICE'] - $bonus_summ;

		if ($order_price > 0 && $new_order_price > 0 && $order_price != $new_order_price) {
			$dif = round(($order_price - $new_order_price), 2);
			foreach ($products as $key=>$product) {
				if ($product->price > 0 && $product->price > $dif) {
					if (
						$product->count == 1 ||
						$product->count == abs($dif)*100 || 
						($product->count % 2 == 0 && (abs($dif)*100) % 2 == 0) ||
						($product->count % 2 == 1 && (abs($dif)*100) % 2 == 1)
					) {
						if (abs($dif) / $product->count >= 0.01) {
							$products[$key]->price += round(($dif / $product->count), 2);
							$correction = 1;
							break;
						}
					}
				}
			}
		}

		$order['PRICE'] = $correction ? $order_price : $new_order_price;

		if ($order_price <= 0) {
			foreach ($products as $key => $product) {
				$products[$key]->price = 0;
			}

			$order['PRICE_DELIVERY'] = 0;
			$order['PRICE'] = 0;
		}

		// Avans
		$is_avans = null;
		$avans = COption::GetOptionString($mid, 'avans');
		$summ_paid = self::__formatprc($order['SUM_PAID'] - $bonus_summ);
		if (
			$avans && 
			$summ_paid > 0 && 
			self::__formatprc($order['PRICE']) > $summ_paid
		) {
			$is_avans = 1;

			$product = new stdClass();
			$product->id = 'avans' . $summ_paid;
			$product->name = 'АВАНС';
			$product->price = $summ_paid;
			$product->count = 1;
			$product->tax = 'WITHOUT_VAT';
			$product->tax_value = 0;
			$product->code = 'avans' . $summ_paid;
			$products = [$product];

			$order_delivery_price = 0;
			$order['PRICE_DELIVERY'] = 0;
			$order['PRICE'] = $summ_paid;
		}

		// delivery
		$delivery = CSaleDelivery::GetByID($order['DELIVERY_ID']);
		$vat_id = $delivery['VAT_ID'];
		$vatList = array();
		if (('\Bitrix\Main\Loader')::includeModule('catalog')) {
			$dbRes = ('\Bitrix\Catalog\VatTable')::getList(array(
				'filter' => array('ACTIVE' => 'Y'),
				'order' => array('SORT' => 'ASC')
			));
			while ($vat = $dbRes->fetch()) {
				$vatList[$vat['ID']] = (int) $vat['RATE'];
			}
		}
		$delivery_tax =
			$vatList[$vat_id] > 0 ? 'VAT_' . $vatList[$vat_id] : 'WITHOUT_VAT';
		$delivery_tax_value = $vatList[$vat_id] > 0 ? $vatList[$vat_id] : 0;

		// data
		$data = new stdClass();
		$data->id = $order['ID'];
		$data->number = $order['ACCOUNT_NUMBER'];
		$data->created = $order['DATE_INSERT'];
		$data->changed = date('Y-m-d H:i:s');
		$data->total = self::__formatprc($order['PRICE']);
		$data->delivery_id = $order['DELIVERY_ID'];
		$data->delivery_cost = $order_delivery_price;
		$data->delivery_tax = $delivery_tax;
		$data->delivery_tax_value = $delivery_tax_value;
		$data->payment_id = $order['PAY_SYSTEM_ID'];
		$data->status_id = $order['STATUS_ID'];
		$data->paid = $order['PAYED'];
		$data->pay_date = $order['DATE_PAYED'];
		$data->sum_paid = self::__formatprc($order['SUM_PAID'] - $bonus_summ);
		$data->payments = self::__getOrderPayments($order['ID']);
		$data->comment = self::__clearString($order['COMMENTS']);
		$data->user_comment = self::__clearString($order['USER_DESCRIPTION']);
		$data->user_name = self::__clearString($order['USER_NAME']);
		$data->user_last_name = self::__clearString($order['USER_LAST_NAME']);
		$data->user_email =
			isset($order['USER_EMAIL']) && $order['USER_EMAIL'] != ''
				? self::__checkEmail($order['USER_EMAIL'])
				: null;
		$data->user_phone =
			isset($user_phone) && $user_phone != ''
				? self::__checkPhone($user_phone)
				: null;
		$data->user_address =
			isset($user_address) && $user_address != '' ? self::__clearString($user_address) : null;
		$data->address =
			$address_type == 'email' ? $data->user_email : $data->user_phone;
		$data->products = $products;
		$data->cashier_name = COption::GetOptionString($mid, 'cashier_name');
		$data->cashier_inn = COption::GetOptionString($mid, 'cashier_inn');
		$data->payment_address = COption::GetOptionString(
			$mid,
			'payment_address'
		);
		$data->calc_method = $is_avans
			? COption::GetOptionString($mid, 'avans_calc_method')
			: COption::GetOptionString($mid, 'calculation_method');
		$data->payment_type = $is_avans 
			? 'CARD' 
			: COption::GetOptionString($mid, 'payment_type');
		$data->fio = COption::GetOptionString($mid, 'fio');
		$data->vat_mode = COption::GetOptionString($mid, 'vat_mode');
		$data->vat = COption::GetOptionString($mid, 'vat');
		$data->vat_value = str_replace(
			['WITHOUT_VAT', 'VAT_'],
			['0', ''],
			strtoupper($data->vat)
		);
		$data->avans = COption::GetOptionString($mid, 'avans');
		$data->avans_calc_method = COption::GetOptionString(
			$mid,
			'avans_calc_method'
		);
		$data->bonus = COption::GetOptionString($mid, 'bonus');
		$data->pm_bonus_ids = json_decode(
			COption::GetOptionString($mid, 'pm_bonus_ids')
		);
		$data->bonus_summ = self::__getPayment($data->id, $data->pm_bonus_ids);
		$data->ok_payments = self::__getOrderPayments($data->id, $data->pm_bonus_ids, 1);

		return $data;
	}

	public static function handleSaleOrderStatus($id, $val)
	{
		if (gettype($id) == 'object' || !CModule::IncludeModule('sale')) {
			return;
		}

		// order data
		$order = CSaleOrder::GetByID($id);
		$mid = self::MODULE_ID;
		$data = self::__getOrderData($order);
		$data->test = COption::GetOptionString($mid, 'test') ? '-test' : '';
		$data->token =
			$data->test != ''
				? 'P5cKbUUD9uSSrSlGdzspLblvBnD0GzTAE0cLmAPSEMxJ79DtLE'
				: COption::GetOptionString($mid, 'token');
		$status_id = $val;
		$payment_id = $data->payment_id;
		$paid = $data->paid;

		// no action duplicates
		$no_duplicates = COption::GetOptionString($mid, 'no_duplicates', 0);
		if ($no_duplicates) {
			$session = ('\Bitrix\Main\Application')
				::getInstance()
				->getSession();
		}

		// iterating over the document type tabs
		for ($i = 1; $i <= 5; $i++) {
			// will find out whether it is necessary to use this type of document
			$use_type = COption::GetOptionString($mid, 'use_type_' . $i)
				? 1
				: null;
			// if need to use
			if ($use_type) {
				// get parameters
				/// order paid (yes / no / disregard) (2/1/0)
				$must_paid = COption::GetOptionString($mid, 'order_paid_' . $i);
				/// selected order status (s)
				$pm_statuses = json_decode(
					COption::GetOptionString($mid, 'pm_statuses_' . $i)
				);
				/// selected payment method (s)
				$pm_ids = json_decode(
					COption::GetOptionString($mid, 'pm_ids_' . $i)
				);
				/// use refused
				$use_refused = COption::GetOptionString($mid, 'lk_use_refused');
				/// connection tab
				if ($i == 5 && $use_refused) {
					//// selected refused status (s)
					$lk_refused_statuses = json_decode(
						COption::GetOptionString($mid, 'lk_refused_statuses')
					);
					/// if in array merge status arrays
					if (in_array($status_id, $lk_refused_statuses)) {
						$pm_statuses = array_merge(
							$pm_statuses,
							$lk_refused_statuses
						);
					}
				}

				// If the payment method and order status match
				if (
					isset($pm_ids) &&
					isset($pm_statuses) &&
					in_array($payment_id, $pm_ids) &&
					in_array($status_id, $pm_statuses) &&
					$data->token != ''
				) {
					// if "order paid" is selected,
					// and it is NOT paid, then we check the following tab
					if ($must_paid == '2' && $paid != 'Y') {
						continue;
					}

					// if "order not paid" is selected,
					// and it is paid, then we check the next tab
					if ($must_paid == '1' && $paid != 'N') {
						continue;
					}

					// check for duplicates
					if ($no_duplicates) {
						$act_name = 'bifit_act_' . $i . '_' . $id;
						if ($session->has($act_name)) {
							if (
								$session->get($act_name) >
								strtotime(date("Y-m-d H:i:s"))
							) {
								continue;
							}
						}
					}

					// update data
					$arr_rec_type = array(
						'1' => 'SALE',
						'2' => 'SALE_RETURN',
						'3' => 'PURCHASE',
						'4' => 'PURCHASE_RETURN'
					);

					$data->receipt_type = $arr_rec_type[$i];
					$data->tax_system = COption::GetOptionString(
						$mid,
						'tax_system_' . $i
					);

					// main processes
					if ($i == 5) {
						// update data
						$data->login = self::__checkPhone(
							COption::GetOptionString($mid, 'login')
						);
						$data->password = COption::GetOptionString(
							$mid,
							'password'
						);
						$data->org_id = COption::GetOptionString(
							$mid,
							'organization'
						);
						$data->trade_id = COption::GetOptionString(
							$mid,
							'trade_object'
						);
						$data->courier = json_decode(
							COption::GetOptionString(
								$mid,
								'lk_courier_delivery'
							)
						);
						$data->self = json_decode(
							COption::GetOptionString($mid, 'lk_self_delivery')
						);
						$data->new_status = COption::GetOptionString(
							$mid,
							'lk_new_courier_order_status'
						);
						$data->use_refused = COption::GetOptionString(
							$mid,
							'lk_use_refused'
						);
						$data->refused_statuses = json_decode(
							COption::GetOptionString(
								$mid,
								'lk_refused_statuses'
							)
						);
						// do
						self::__saveOrder($data);
					} else {
						// do
						self::__doFiscalize($data);
					}

					// save to session and check old values
					if ($no_duplicates) {
						if (!$session->has('bifit_acts')) {
							$session['bifit_acts'] = [];
						} else {
							$arr = $session->get('bifit_acts');
							if (is_array($arr) && count($arr)) {
								foreach ($arr as $k => $v) {
									$a_n = 'bifit_act_' . $v;
									if ($session->has($a_n)) {
										if (
											strtotime(date("Y-m-d H:i:s")) >
											$session->get($a_n)
										) {
											$session->remove($a_n);
											unset($session['bifit_acts'][$k]);
										}
									}
								}
							}
						}
						$session['bifit_acts'][] = $i . '_' . $id;
						$session->set(
							$act_name,
							strtotime(date("Y-m-d H:i:s") . "+5 seconds")
						);
					}
				}
			}
		}
	}

	public static function handleSaleOrderPay($id, $val)
	{
		if (gettype($id) == 'object' || !CModule::IncludeModule('sale')) {
			return;
		}

		// order data
		$order = CSaleOrder::GetByID($id);
		$mid = self::MODULE_ID;
		$data = self::__getOrderData($order);
		$data->test = COption::GetOptionString($mid, 'test') ? '-test' : '';
		$data->token =
			$data->test != ''
				? 'P5cKbUUD9uSSrSlGdzspLblvBnD0GzTAE0cLmAPSEMxJ79DtLE'
				: COption::GetOptionString($mid, 'token');
		$status_id = $data->status_id;
		$payment_id = $data->payment_id;
		$paid = $val;

		// no action duplicates
		$no_duplicates = COption::GetOptionString($mid, 'no_duplicates', 0);
		if ($no_duplicates) {
			$session = ('\Bitrix\Main\Application')
				::getInstance()
				->getSession();
		}

		// iterating over the document type tabs
		for ($i = 1; $i <= 5; $i++) {
			// will find out whether it is necessary to use this type of document
			$use_type = COption::GetOptionString($mid, 'use_type_' . $i)
				? 1
				: null;
			// if need to use
			if ($use_type) {
				// get parameters
				/// order paid (yes / no / disregard) (2/1/0)
				$must_paid = COption::GetOptionString($mid, 'order_paid_' . $i);
				/// selected order status (s)
				$pm_statuses = json_decode(
					COption::GetOptionString($mid, 'pm_statuses_' . $i)
				);
				/// selected payment method (s)
				$pm_ids = json_decode(
					COption::GetOptionString($mid, 'pm_ids_' . $i)
				);

				// If the payment method and order status match
				if (
					isset($pm_ids) &&
					isset($pm_statuses) &&
					in_array($payment_id, $pm_ids) &&
					in_array($status_id, $pm_statuses) &&
					$data->token != ''
				) {
					// if "order paid" is selected,
					// and it is NOT paid, then we check the following tab
					if ($must_paid == '2' && $paid != 'Y') {
						continue;
					}

					// if "order not paid" is selected,
					// and it is paid, then we check the next tab
					if ($must_paid == '1' && $paid != 'N') {
						continue;
					}

					// check for duplicates
					if ($no_duplicates) {
						$act_name = 'bifit_act_' . $i . '_' . $id;
						if ($session->has($act_name)) {
							if (
								$session->get($act_name) >
								strtotime(date("Y-m-d H:i:s"))
							) {
								continue;
							}
						}
					}

					// update data
					$arr_rec_type = array(
						'1' => 'SALE',
						'2' => 'SALE_RETURN',
						'3' => 'PURCHASE',
						'4' => 'PURCHASE_RETURN'
					);

					$data->receipt_type = $arr_rec_type[$i];
					$data->tax_system = COption::GetOptionString(
						$mid,
						'tax_system_' . $i
					);

					// main processes
					if ($i == 5) {
						// update data
						$data->login = self::__checkPhone(
							COption::GetOptionString($mid, 'login')
						);
						$data->password = COption::GetOptionString(
							$mid,
							'password'
						);
						$data->org_id = COption::GetOptionString(
							$mid,
							'organization'
						);
						$data->trade_id = COption::GetOptionString(
							$mid,
							'trade_object'
						);
						$data->courier = json_decode(
							COption::GetOptionString(
								$mid,
								'lk_courier_delivery'
							)
						);
						$data->self = json_decode(
							COption::GetOptionString($mid, 'lk_self_delivery')
						);
						$data->new_status = COption::GetOptionString(
							$mid,
							'lk_new_courier_order_status'
						);
						$data->use_refused = 0;
						$data->refused_statuses = [];
						// do
						self::__saveOrder($data);
					} else {
						// do
						self::__doFiscalize($data);
					}

					// save to session and check old values
					if ($no_duplicates) {
						if (!$session->has('bifit_acts')) {
							$session['bifit_acts'] = [];
						} else {
							$arr = $session->get('bifit_acts');
							if (is_array($arr) && count($arr)) {
								foreach ($arr as $k => $v) {
									$a_n = 'bifit_act_' . $v;
									if ($session->has($a_n)) {
										if (
											strtotime(date("Y-m-d H:i:s")) >
											$session->get($a_n)
										) {
											$session->remove($a_n);
											unset($session['bifit_acts'][$k]);
										}
									}
								}
							}
						}
						$session['bifit_acts'][] = $i . '_' . $id;
						$session->set(
							$act_name,
							strtotime(date("Y-m-d H:i:s") . "+5 seconds")
						);
					}
				}
			}
		}
	}
}
